' this selects the mailbox interface
'#define USE_JCACHED_MEMORY

' select only one of the below
'#define USE_SDRAM_MEMORY        ' John Steven Denson (jazzed)'s SDRAM support
'#define USE_C3_SPI_MEMORY       ' Parallax's C3 SPI flash/SRAM support

' load the serial debug stub
'#define SERIAL_DEBUG

' select debugging output to a TV
#define TV_DEBUG

#ifdef USE_SDRAM_MEMORY
#define XCACHE
#endif

CON
  _clkmode = def#_clkmode
  _xinfreq = def#_xinfreq

  TYPE_FILE_WRITE = 1
  TYPE_FLASH_WRITE = 2
  TYPE_RAM_WRITE = 3
  TYPE_DATA = 4
  TYPE_EOF = 5
  TYPE_FILE_LOAD = 6
  TYPE_RUN = 7

  HDR_SOH = 0
  HDR_PKTN = 1
  HDR_TYPE = 2
  HDR_LEN_HI = 3
  HDR_LEN_LO = 4
  HDR_CHK = 5

  ' character codes
  SOH = $01    ' start of a packet
  ACK = $06
  NAK = $15
  CR  = $0d

  ' serial error codes
  SERIAL_ERROR = -1
  SERIAL_TIMEOUT = -2

  WRITE_FILE = 1
  WRITE_FLASH = 2
  WRITE_RAM = 3

  hub_memory_size = 32 * 1024
  rendezvous_area_size = 128
  zog_dispatch_table_size = 128

#ifdef USE_C3_SPI_MEMORY
  zpu_memory_size = 64 * 1024
#endif
#ifdef USE_SDRAM_MEMORY
  zpu_memory_size = 32 * 1024 * 1024
#endif
#ifdef USE_DRACBLADE_MEMORY
  zpu_memory_size = 512 * 1024
#endif
  vm_cache_size = 8 * 1024
  vm_mbox_size = 2 * 4
  zog_mbox_size = 8 * 4
  debug_control_size = 10 * 4
  debug_buffer_size = 16 * 2
  rendezvous_area = hub_memory_size - rendezvous_area_size
  zog_dispatch_table = rendezvous_area - zog_dispatch_table_size
  vm_cache = zog_dispatch_table - vm_cache_size
  vm_mbox = rendezvous_area
  zog_mbox = vm_mbox + vm_mbox_size
  debug_control = zog_mbox + zog_mbox_size
  debug_buffer = debug_control + debug_control_size

OBJ
#ifdef USE_C3_SPI_MEMORY
  def  : "userdefs_c3.spin"
  vm   : "spisram_cache_c3"
#endif
#ifdef USE_SDRAM_MEMORY
  def  : "userdefs_sdram.spin"
  vm   : "SdramCache"
#endif
#ifdef USE_DRACBLADE_MEMORY
  def  : "userdefs_dracblade.spin"
  vm   : "sram_cache_dracblade"
#endif
  zog  : "zog.spin"
  pkt  : "packet_driver"
  sd   : "fsrwFemto_rr001"
  dbg  : "serial_debug"
#ifdef TV_DEBUG
  tv   : "TV_Text"
#endif

VAR
  '8 consecutive longs for Zog's par block.
  long par_zpu_memory_addr
  long par_zpu_memory_sz
  long par_initial_pc
  long par_initial_sp
  long par_dispatch_tab_addr
  long par_pasm_addr
  long par_zog_mbox_addr
  long par_vm_mbox_addr

  long ioControl[2]      ' SD parameters
  byte diskbuff[512]     ' SD sector buffer
  byte filenamebuff[129] ' Image filename
  long run_ram_image

  long sd_mounted
  long write_mode
  long load_address

PUB start | type, packet, len, timeout, ok

  pkt.start(def#conRxPin, def#conTxPin, def#conMode, def#conBaud)

#ifdef TV_DEBUG
  tv.start(def#TvPin)
#ifdef SERIAL_DEBUG
  tv.str(string("Serial Debug Loader v0.3 ("))
#else
  tv.str(string("Serial Loader v0.3 ("))
#endif
  tv.str(@board_name)
  tv.str(string(")", CR))
#endif

  vm.startx(vm_mbox, vm_cache)

  run_ram_image := 0
  sd_mounted := 0

  ' use a one second timeout for the first packet
  ' if no packet is received before the timeout,
  ' start the program whose name is in the SD card
  ' file "zogload.run"
  timeout := 1000

  repeat

    ' get the next packet from the PC
    ok := pkt.rcv_packet(@type, @packet, @len)
#ifdef TV_DEBUG
    if not ok
      tv.str(string("Receive packet error", CR))
#endif

    ' check for timeout on the first packet
    if len == SERIAL_TIMEOUT
      start_zogload_run_program
    timeout := 0

    if ok
      case type
        TYPE_FILE_WRITE:        FILE_WRITE_handler(packet)
        TYPE_RAM_WRITE:         RAM_WRITE_handler
        TYPE_DATA:              DATA_handler(packet, len)
        TYPE_EOF:               EOF_handler
        TYPE_FILE_LOAD:         FILE_LOAD_handler(packet)
        TYPE_RUN:               RUN_handler
#ifdef USE_C3_SPI_MEMORY
        TYPE_FLASH_WRITE:       FLASH_WRITE_handler
#endif
        other:
#ifdef TV_DEBUG
          tv.str(string("Bad packet type: "))
          tv.hex(type, 2)
          crlf
#endif
      pkt.release_packet

PRI FILE_WRITE_handler(name) | err
  mountSD
#ifdef TV_DEBUG
  tv.str(string("FILE_WRITE: "))
  tv.str(name)
  tv.str(string("...",))
#endif
  err := \sd.popen(name, "w")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
#endif
  write_mode := WRITE_FILE
  load_address := $00000000

PRI RAM_WRITE_handler
#ifdef TV_DEBUG
  tv.str(string("RAM_WRITE", CR))
#endif
  write_mode := WRITE_RAM
  load_address := $00000000
  run_ram_image := 1

PRI DATA_handler(packet, len) | i, val
#ifdef TV_DEBUG
  tv.str(string("DATA: "))
  tv.hex(load_address, 8)
  tv.out(" ")
  tv.dec(len)
  crlf
#endif
  case write_mode
    WRITE_FILE:
      sd.pwrite(packet, len)
      load_address += len
    WRITE_RAM:
      repeat i from 0 to len - 1 step 4
        val := (byte[packet][i] << 24) | (byte[packet][i+1] << 16) | (byte[packet][i+2] << 8) | byte[packet][i+3]
        vm.writeLong(load_address, val)
        load_address += 4
#ifdef USE_C3_SPI_MEMORY
    WRITE_FLASH:
      vm.WriteFlash(load_address, packet, len)
      load_address += len
      if (load_address & $00000fff) == 0
        vm.eraseFlashBlock(load_address)
#endif
    other:
#ifdef TV_DEBUG
      tv.str(string("Bad write_mode: "))
      tv.hex(write_mode, 8)
      crlf
#endif

PRI EOF_handler
#ifdef TV_DEBUG
  tv.str(string("EOF", CR))
#endif
  case write_mode
    WRITE_FILE:
      sd.pclose
    WRITE_RAM:
#ifdef USE_C3_SPI_MEMORY
    WRITE_FLASH:
#endif
    other:
#ifdef TV_DEBUG
      tv.str(string("Bad write_mode: "))
      tv.hex(write_mode, 8)
      crlf
#endif

PRI FILE_LOAD_handler(name)
#ifdef TV_DEBUG
  tv.str(string("FILE_LOAD: "))
  tv.str(name)
  crlf
#endif
  mountSD
  load_image(name)

PRI RUN_handler
#ifdef TV_DEBUG
  tv.str(string("RAM_RUN", CR))
#endif
  if  run_ram_image
    start_bytecode_in_ram
  else
#ifdef USE_C3_SPI_MEMORY
    start_bytecode_in_flash
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash memory not supported"))
#endif
  start_zog

#ifdef USE_C3_SPI_MEMORY

PRI FLASH_WRITE_handler
#ifdef TV_DEBUG
  tv.str(string("FLASH_WRITE", CR))
#endif
  write_mode := WRITE_FLASH
  load_address := $00000000 ' offset into flash
  vm.eraseFlashBlock(load_address)
  run_ram_image := 0

#endif

#ifdef TV_DEBUG
PRI crlf
  tv.out(CR)
#endif

PRI mountSD | err
  if sd_mounted == 0
    if sd.startSD(@ioControl) == 0
      return 0
    repeat
#ifdef TV_DEBUG
      tv.str(string("Mounting SD card...", CR))
#endif
      err := \sd.mount_explicit(def#spiDO,def#spiClk,def#spiDI,def#spiCS)
    until err == 0
    sd_mounted := 1
  return 1

PUB start_zogload_run_program
  mountSD
  read_zogload_run(@filenamebuff)
  load_image(@filenamebuff)
  if run_ram_image
    start_bytecode_in_ram
  else
#ifdef USE_C3_SPI_MEMORY
    start_bytecode_in_flash
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash memory not supported"))
#endif
  start_zog

PUB read_zogload_run(filename) | bytes, err
#ifdef TV_DEBUG
  tv.str(string("Opening 'zogload.run'..."))
#endif
  err := \sd.popen(string("zogload.run"), "r")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
#endif
  if err <> 0
    Fatal(string("Can't open file"))
#ifdef TV_DEBUG
  tv.str(string("Reading filename..."))
#endif
  bytes := sd.pread(filename, 128)
  CheckError(bytes)
  byte[filename][bytes] := 0
#ifdef TV_DEBUG
  tv.str(filename)
  crlf
#endif
  err := sd.pclose
  CheckError(err)

PUB load_image(filename) | bytes, err
#ifdef TV_DEBUG
  tv.str(string("Opening ZPU image: "))
  tv.str(filename)
  tv.str(string("..."))
#endif
  err := \sd.popen(filename, "r")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
#endif
  if err <> 0
    Fatal(string("Can't open file"))
#ifdef TV_DEBUG
  tv.str(string("Reading image..."))
#endif

  bytes := sd.pread(@diskbuff, 512)
  if bytes < 4
    Fatal(string("Bad image header"))

  if long[@diskbuff] == $00000000
    load_bytecode(bytes)
  else
#ifdef USE_C3_SPI_MEMORY
    write_bytecode_to_flash(bytes)
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash loads not supported"))
#endif

' called with data already in diskbuff
PUB load_bytecode(bytes) | addr, val, i, err
  addr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  CheckError(err)
  repeat while bytes > 0
    repeat i from 0 to bytes - 1 step 4
      val := (diskbuff[i] << 24) | (diskbuff[i+1] << 16) | (diskbuff[i+2] << 8) | diskbuff[i+3]
      vm.writeLong(addr, val)
      addr += 4
    err := \sd.initSDCard(def#spiDO,def#spiClk,def#spiDI,def#spiCS)     'Init the SD
    bytes := sd.pread(@diskbuff, 512)
    err := sd.stopSDCard                                'Stop SD card, but do not stop cog
    CheckError(err)
#ifdef TV_DEBUG
  tv.dec(addr)
  tv.str(string(" bytes"))
  crlf
#endif
  run_ram_image := 1

PUB start_bytecode_in_ram | bss_start, bss_end, i

  ' load the image header
  load_address   := swap(vm.readLong($00000000))
  par_initial_pc := swap(vm.readLong($00000004))
  bss_start      := swap(vm.readLong($00000008))
  bss_end        := swap(vm.readLong($0000000c))

#ifdef TV_DEBUG
  taghex8(string("load_address: "), load_address)
  taghex8(string("start:        "), par_initial_pc)
  taghex8(string("bss_start:    "), bss_start)
  taghex8(string("bss_end:      "), bss_end)
#endif

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

#ifdef USE_C3_SPI_MEMORY

' called with data already in diskbuff
PUB write_bytecode_to_flash(bytes) | i, addr, err
  addr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  CheckError(err)
  vm.eraseFlashBlock(addr)
  repeat while bytes > 0
    if bytes > 0
      vm.writeFlash(addr, @diskbuff, bytes)
      addr += bytes
      if (addr & $00000fff) == 0
        vm.eraseFlashBlock(addr)
      err := \sd.initSDCard(def#spiDO,def#spiClk,def#spiDI,def#spiCS)     'Init the SD
      bytes := sd.pread(@diskbuff, 512)
      err := sd.stopSDCard                                'Stop SD card, but do not stop cog
      CheckError(err)
#ifdef TV_DEBUG
  tv.dec(addr)
  tv.str(string(" bytes"))
  crlf
#endif
  run_ram_image := 0

PUB start_bytecode_in_flash | bss_start, bss_end, data_image, data_start, data_end

  ' load the image header
  load_address   := swap(vm.readLong($00100000))
  par_initial_pc := swap(vm.readLong($00100004))
  bss_start      := swap(vm.readLong($00100008))
  bss_end        := swap(vm.readLong($0010000c))
  data_image     := swap(vm.readLong($00100010))
  data_start     := swap(vm.readLong($00100014))
  data_end       := swap(vm.readLong($00100018))

#ifdef TV_DEBUG
  taghex8(string("load_address: "), load_address)
  taghex8(string("start:        "), par_initial_pc)
  taghex8(string("bss_start:    "), bss_start)
  taghex8(string("bss_end:      "), bss_end)
  taghex8(string("data_image:   "), data_image)
  taghex8(string("data_start:   "), data_start)
  taghex8(string("data_end:     "), data_end)
#endif

  ' initialize .data
  repeat while data_start < data_end
    vm.writeLong(data_start, vm.readLong(data_image))
    data_image += 4
    data_start += 4

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

#endif

PRI swap(value) | swapped
  return value
  swapped :=  value              << 24
  swapped |= (value & $0000ff00) << 8
  swapped |= (value & $00ff0000) >> 8
  swapped |=  value              >> 24
  return swapped

#ifdef TV_DEBUG
PRI taghex8(tag, addr)
  tv.str(tag)
  tv.hex(addr, 8)
  crlf
#endif

PUB start_zog | cog

  'Set up Zog's parameter block (except par_initial_pc)
  par_zpu_memory_addr := 0      ' using external memory
  par_zpu_memory_sz := 0        ' using external memory
  par_initial_sp := zpu_memory_size - 8
  par_dispatch_tab_addr := zog.getdispatch_table
  par_pasm_addr := zog.getzog
  par_zog_mbox_addr := zog_mbox
  par_vm_mbox_addr := vm_mbox

  waitcnt(clkfreq+cnt)                                'Give serial terminal window time

  ' stop all COGs except the one running the vm
  pkt.stop
  sd.stop

#ifdef SERIAL_DEBUG
  ' start the serial debug COG
  cog := dbg.start(debug_control, def#conRxPin, def#conTxPin, def#conMode, def#conBaud, debug_buffer, zog_mbox)
#ifdef TV_DEBUG
  if cog == 0
    tv.str(string("Failed to start serial debug COG", CR))
    repeat
#endif
#endif

#ifdef TV_DEBUG
  tv.stop
#endif

  ' let ZOG replace the Spin loader code
  coginit(cogid, @boot_zog, @par_zpu_memory_addr)

PRI Fatal(msg)
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.str(msg)
    crlf
#endif
    repeat                                             '<== LOOP HERE

PRI CheckError(r)
  if r < 0
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.hex(r,2)
    tv.str(string(" HALTED."))
    crlf
#endif
    repeat                                             '<== LOOP HERE

DAT
                        org     0
'Copy PAR block into COG
boot_zog                mov     zog_params, par
                        mov     address, zog_params
                        add     address, #16
                        rdlong  dispatch_tab_addr, address
                        wrlong  dispatch_tab_final, address 'Update param block with new address
                        add     address, #4
                        rdlong  pasm_addr, address

'Move ZOG dispatch table runtime position in HUB RAM
move_dispatch           mov     src, dispatch_tab_addr
                        mov     dest, dispatch_tab_final
                        mov     count, #$80
:loop                   rdbyte  temp, src
                        wrbyte  temp, dest
                        add     src, #1
                        add     dest, #1
                        djnz    count, #:loop

'Launch Zog in this COG.
go_zog                  cogid   coginit_dest            'Get ID of this COG for restarting
                        mov     temp, pasm_addr         'Set ZOG PASM address for coginit
                        shl     temp, #2                'Shift into bits 17:4 (only high 14 bits required)
                        or      coginit_dest, temp      'Place in dest for coginit

                        mov     temp, zog_params        'Set PAR block address for coginit
                        shl     temp, #16               'Move to bits 31:18 (only high 14 bits required)
                        or      coginit_dest, temp      'Combine PASM addr and PAR addr

                        coginit coginit_dest

#ifdef USE_C3_SPI_MEMORY
board_name              byte    "C3", 0
#endif
#ifdef USE_SDRAM_MEMORY
board_name              byte    "SDRAM", 0
#endif
#ifdef USE_DRACBLADE_MEMORY
board_name              byte    "DRACBLADE", 0
#endif

zog_params              long    0
dispatch_tab_addr       long    0
pasm_addr               long    0
address                 long    0
dest                    long    0
src                     long    0
count                   long    0
temp                    long    0
coginit_dest            long    0

'The final run time position of Zog's dispatch table
dispatch_tab_final      long    zog_dispatch_table

                        fit     $01ff

'The End.

